from .stadium2d import Stadium2D
from .stadium3d import Stadium3D


class OlympicVelodrome2D(Stadium2D):
    """
    Wrapper class for an Olympic Velodrome in 2D.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            **kwargs,
            length=250,
            radius=24.37,
            width=8,
            straight_banking=12,
            curve_banking=45,
            lane_widths=[0.7, 3.3],
            lane_colors=["black", "red", "blue"],
            surface_color="peru",
            surface_alpha=0.5,
            infield_width=2,
            infield_color="blue",
            infield_alpha=0.3,
        )


class OlympicVelodrome3D(Stadium3D):
    """
    Wrapper class for an Olympic Velodrome in 3D.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            **kwargs,
            length=250,
            radius=24.37,
            width=8,
            straight_banking=12,
            curve_banking=45,
            lane_widths=[0.7, 3.3],
            lane_colors=["black", "red", "blue"],
            surface_color="peru",
            surface_alpha=0.5,
            infield_width=2,
            infield_color="blue",
            infield_alpha=0.3,
        )


class OutdoorAthleticsTrack2D(Stadium2D):
    """
    Wrapper class for an Outdoor Athletics Track in 2D.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            **kwargs,
            length=400,
            radius=36.5,
            width=10,
            straight_banking=0,
            curve_banking=0,
            lane_widths=[1.25] * 8,
            lane_colors=["white"] * 9,
            line_distances=[42.66, 242.66],
            line_colors=["white"] * 2,
        )


class OutdoorAthleticsTrack3D(Stadium3D):
    """
    Wrapper class for an Outdoor Athletics Track in 3D.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            **kwargs,
            length=400,
            radius=36.5,
            width=10,
            straight_banking=0,
            curve_banking=0,
            lane_widths=[1.25] * 8,
            lane_colors=["white"] * 9,
            line_distances=[42.66, 242.66],
            line_colors=["white"] * 2,
        )


class IndoorAthleticsTrack2D(Stadium2D):
    """
    Wrapper class for an Indoor Athletics Track in 2D.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            **kwargs,
            length=200,
            radius=18.5,
            width=7.5,
            straight_banking=0,
            curve_banking=10,
            lane_widths=[1.25] * 6,
            lane_colors=["white"] * 7,
        )


class IndoorAthleticsTrack3D(Stadium3D):
    """
    Wrapper class for an Indoor Athletics Track in 3D.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            **kwargs,
            length=200,
            radius=18.5,
            width=7.5,
            straight_banking=0,
            curve_banking=10,
            lane_widths=[1.25] * 6,
            lane_colors=["white"] * 7,
        )



import abc
from typing import List, Tuple, Union

import matplotlib
import numpy as np
from matplotlib import pyplot as plt


class StadiumBase(abc.ABC):
    """

    An abstract base class to represent a stadium-shaped track and facilitate plots of and on its surface.

    Attributes:
    ----------
    length : float
        The length of the inner line of the stadium.
    radius : float
        The radius of the corners of the stadium.
    width : float
        The width of the stadium.
    straight_banking : float
        The banking angle (deg) of the straight sections of the stadium.
    curve_banking : float
        The banking angle (deg) of the curved sections of the stadium.
    lane_widths : Iterable[float] = [1]
        Optional. Widths for each lane of the stadium. Does not have to sum to the total width
    lane_colors : Iterable[str] = ["black", "black"],
        Optional. Lane line colours, including the inner and outer lane lines.
        Must have length equal to len(lane_widths) + 1
    lane_alpha : float = 1.0
        Optional. Alpha value of the lane lines.
    lane_linestyle : str = "-"
        Optional. Linestyle of the lane lines.
    lane_linewidth : float
        Optional. Linewidth of the lane lines.
    line_distances : Iterable[float] = []
        Optional. Distances of the radial lines on the stadium.
    line_colors : Iterable[str] = []
        Optional. Colors of the distance lines. Must be of equal length to line_distances.
    line_alpha : float = 1.0
        Optional. Alpha value of the radial distance lines
    surface_color : str = "blue"
        Optional. Surface color of the stadium
    surface_alpha : float = 0.5
        Optional. Surface alpha of the stadium.
        Note that alpha near one will mean lines and scatters are difficult to see.
    infield_color : str = "white"
        Optional. Color of the infield area within the stadium.
    infield_alpha : str
        Optional. Alpha of the infield area.
    infield_width : float | "all" = "all"
        Optional. Width of the infield, or "all" for full infield area.

    """

    def __init__(
        self,
        length: float,
        radius: float,
        width: float,
        straight_banking: float,
        curve_banking: float,
        lane_widths: tuple[float] = (1,),
        lane_colors: tuple[str] = ("black", "black"),
        lane_alpha: float = 1,
        lane_linestyle: str = "-",
        lane_linewidth: float = 1,
        line_distances: tuple[float] = (),
        line_colors: tuple[str] = (),
        line_alpha: float = 1,
        surface_color: str = "blue",
        surface_alpha: float = 1,
        infield_color: str = "white",
        infield_alpha: float = 1,
        infield_width: Union[float, str] = "all",
    ):
        assert len(lane_colors) == len(lane_widths) + 1, "Must have one more lane colors than lane widths"
        assert len(line_distances) == len(line_colors), "Must have same number of line distances and line colors"
        assert isinstance(infield_width, (float, int)) or infield_width == "all", 'Infield width must be float or "all"'

        self.length = length
        self.radius = radius
        self.width = width
        self.straight_banking = straight_banking
        self.curve_banking = curve_banking
        self.lane_widths = lane_widths
        self.lane_colors = lane_colors
        self.lane_alpha = lane_alpha
        self.lane_linestyle = lane_linestyle
        self.lane_linewidth = lane_linewidth
        self.line_distances = line_distances
        self.line_colors = line_colors
        self.line_alpha = line_alpha
        self.surface_color = surface_color
        self.surface_alpha = surface_alpha
        self.infield_color = infield_color
        self.infield_alpha = infield_alpha
        self.infield_width = infield_width

        self._q_straight: float = (self.length - 2 * np.pi * self.radius) / 4
        self._ax: plt.Axes = None
        self._fig: plt.Figure = None

    def _step_section(
        self,
        start: float,
        end: float,
        f: callable,
    ):
        def _step(s, d_xy, d_z):
            return np.heaviside(s - start, 1) * f(s, d_xy, d_z) - np.heaviside(s - end, 1) * f(s, d_xy, d_z)

        return _step

    def _banking(self, s, d) -> float:
        multiplier = np.heaviside(d, 1)
        return (
            multiplier
            * np.pi
            * (
                ((self.straight_banking + self.curve_banking) / 2)
                - (self.curve_banking - self.straight_banking) / 2 * np.cos(4 * (s / self.length) * np.pi)
            )
            / 180
        )

    def _straight_1(self, s, d_xy, d_z) -> Tuple[float, float, float]:
        return np.array([s, -1 * (self.radius + d_xy), d_z])

    def _curve_1(self, s, d_xy, d_z) -> Tuple[float, float, float]:
        angle = (s - self._q_straight) / self.radius
        return np.array(
            [self._q_straight + (self.radius + d_xy) * np.sin(angle), -1 * (self.radius + d_xy) * np.cos(angle), d_z]
        )

    def straight_2(self, s, d_xy, d_z) -> Tuple[float, float, float]:
        return np.array([2 * self._q_straight + np.pi * self.radius - s, self.radius + d_xy, d_z])

    def curve_2(self, s, d_xy, d_z) -> Tuple[float, float, float]:
        angle = (s - (3 * self._q_straight + np.pi * self.radius)) / self.radius
        return np.array(
            [-1 * self._q_straight - (self.radius + d_xy) * np.sin(angle), (self.radius + d_xy) * np.cos(angle), d_z]
        )

    def straight_3(self, s, d_xy, d_z) -> Tuple[float, float, float]:
        return np.array([s - 4 * self._q_straight - 2 * np.pi * self.radius, -1 * (self.radius + d_xy), d_z])

    def _transform_xyz(self, s, d):
        s = s % self.length
        banking_angle = self._banking(s, d)
        d_xy = d * np.cos(banking_angle)
        d_z = d * np.sin(banking_angle)

        steps = [
            0,
            self._q_straight,
            self._q_straight + np.pi * self.radius,
            3 * self._q_straight + np.pi * self.radius,
            3 * self._q_straight + 2 * np.pi * self.radius,
            self.length,
        ]

        section_functions = [self._straight_1, self._curve_1, self.straight_2, self.curve_2, self.straight_3]

        sections = [self._step_section(steps[i], steps[i + 1], section_functions[i]) for i, _ in enumerate(steps[:-1])]

        return np.sum([section(s, d_xy, d_z) for section in sections], axis=0)

    @abc.abstractmethod
    def _init_ax(self, ax: plt.Axes):
        pass

    @abc.abstractmethod
    def draw(self) -> Tuple[plt.Figure, plt.Axes]:
        pass

    @abc.abstractmethod
    def _draw_stadium(self):
        pass

    @abc.abstractmethod
    def _draw_lanes(self) -> List[matplotlib.lines.Line2D]:
        pass

    @abc.abstractmethod
    def _draw_infield(self) -> List[matplotlib.patches.Polygon]:
        pass

    @abc.abstractmethod
    def _draw_lines(self) -> List[matplotlib.lines.Line2D]:
        pass

    @abc.abstractmethod
    def trajectory(self) -> matplotlib.lines.Line2D:
        pass

    @abc.abstractmethod
    def scatter(self):
        pass



from typing import List, Optional, Tuple

import matplotlib
import numpy as np
from matplotlib import pyplot as plt

from .stadium_base import StadiumBase


class Stadium2D(StadiumBase):
    """
    A class to represent a stadium-shaped track and facilitate plots of and on its surface in 2D.
    """

    def _init_ax(self, ax: plt.Axes):
        self._ax = ax
        if self._ax is None:
            self._fig = plt.figure()
            self._ax = self._fig.add_subplot(111)

    def draw(
        self,
        ax: plt.Axes = None,
        s_points: int = 250,
    ) -> Optional[Tuple[plt.Figure, plt.Axes]]:
        """Plot the stadium in 2D.

        Parameters
        ----------
        ax : plt.Axes
            The axis to plot on.
        s_points: int
            Number of points to plot in the tangential direction

        """

        self._init_ax(ax)

        self._draw_stadium(s_points)
        self._draw_infield(s_points)
        self._draw_lanes()
        self._draw_lines()

        if self._fig:
            return self._fig, self._ax

    def _draw_stadium(
        self,
        s_points=250,
    ):
        all_s = np.linspace(0, self.length, s_points)

        outer = np.array([self._transform_xyz(s, self.width) for s in all_s])
        inner = np.array([self._transform_xyz(s, 0) for s in all_s])

        self._ax.fill(outer[:, 0], outer[:, 1], color=self.surface_color, alpha=self.surface_alpha)
        self._ax.fill(inner[:, 0], inner[:, 1], color="white")

    def _draw_infield(self, s_points=250):
        all_s = np.linspace(0, self.length, s_points)

        outer = np.array([self._transform_xyz(s, 0) for s in all_s])
        self._ax.fill(outer[:, 0], outer[:, 1], color=self.infield_color, alpha=self.infield_alpha)

        if self.infield_width != "all":
            inner = np.array([self._transform_xyz(s, -1 * self.infield_width) for s in all_s])
            self._ax.fill(inner[:, 0], inner[:, 1], color="white")

    def _draw_lanes(self, s_points=250) -> List:
        all_s = np.linspace(0, self.length, s_points)

        lane_positions = [0]
        for w in self.lane_widths:
            lane_positions.append(lane_positions[-1] + w)

        for d, color in zip(lane_positions, self.lane_colors):
            line = np.array([self._transform_xyz(s, d) for s in all_s])
            self._ax.plot(
                line[:, 0],
                line[:, 1],
                color=color,
                alpha=self.lane_alpha,
                linestyle=self.lane_linestyle,
                lw=self.lane_linewidth,
            )

    def _draw_lines(self) -> List:
        for s, color in zip(self.line_distances, self.line_colors):
            line = np.array([self._transform_xyz(s, d) for d in [0, self.width]])
            self._ax.plot(
                line[:, 0],
                line[:, 1],
                color=color,
                alpha=self.line_alpha,
            )

    def trajectory(
        self,
        s_: np.ndarray,
        d_: np.ndarray,
        *args,
        **kwargs,
    ) -> matplotlib.lines.Line2D:
        """Plot a trajectory on the stadium.

        Parameters
        ----------
        s_ : np.ndarray
            The tangential position of the trajectory.
        d_ : np.ndarray
            The radial of the trajectory.
        args : list
            Additional positional arguments to pass to the plot
            function.
        kwargs : dict
            Additional keyword arguments to pass to the plot
            function.

        """
        points = np.array([self._transform_xyz(s_i, d_i) for s_i, d_i in zip(s_, d_)])

        return self._ax.plot(points[:, 0], points[:, 1], *args, **kwargs)

    def scatter(
        self,
        s_: np.ndarray,
        d_: np.ndarray,
        *args,
        **kwargs,
    ) -> matplotlib.collections.PathCollection:
        """Scatter points on the stadium.

        Parameters
        ----------
        s_ : np.ndarray
            The tangential position of the points.
        d_ : np.ndarray
            The radial of the points.
        args : list
            Additional positional arguments to pass to the scatter
            function.
        kwargs : dict
            Additional keyword arguments to pass to the scatter
            function.

        """
        points = np.array([self._transform_xyz(s_i, d_i) for s_i, d_i in zip(s_, d_)])

        return self._ax.scatter(points[:, 0], points[:, 1], *args, **kwargs)



from typing import List, Optional, Tuple

import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from .stadium_base import StadiumBase


class Stadium3D(StadiumBase):
    """
    A class to represent a stadium-shaped track and facilitate plots of and on its surface in 3D.
    """

    def _init_ax(self, ax: Axes3D):
        self._ax = ax
        if self._ax is None:
            self._fig = plt.figure()
            self._ax = self._fig.add_subplot(111, projection="3d", computed_zorder=False)

    def draw(
        self,
        ax: plt.Axes = None,
        s_points: int = 250,
    ) -> Optional[Tuple[plt.Figure, Axes3D]]:
        """Plot the stadium in 3D.

        Parameters
        ----------
        ax : plt.Axes
            The axis to plot on.
        s_points : int
            The number of points to use in the tangential direction.

        """
        self._init_ax(ax)

        self._draw_stadium(s_points)
        self._draw_infield(s_points)
        self._draw_lanes()
        self._draw_lines()

        if self._fig:
            return self._fig, self._ax

    def _draw_stadium(
        self,
        s_points=250,
    ):
        s = np.linspace(0, self.length, s_points)
        d = np.linspace(0, self.width, 3)

        s, d = np.meshgrid(s, d)
        s, d = s.flatten(), d.flatten()

        tri = matplotlib.tri.Triangulation(s, d)

        x, y, z = self._transform_xyz(s, d)

        self._ax.plot_trisurf(
            x,
            y,
            z,
            triangles=tri.triangles,
            alpha=self.surface_alpha,
            color=self.surface_color,
            edgecolor=self.surface_color,
            zorder=-2,
        )

    def _draw_infield(self, s_points=250):
        s = np.linspace(0, self.length, s_points)
        if self.infield_width == "all":
            points = np.array(
                [
                    [(0, 0, 0) for _ in s],
                    [self._transform_xyz(s_, 0) for s_ in s],
                ]
            )
            self._ax.plot_surface(
                points[:, :, 0],
                points[:, :, 1],
                points[:, :, 2],
                color=self.infield_color,
                alpha=self.infield_alpha,
                edgecolors=matplotlib.colors.to_rgba(self.infield_color, self.infield_alpha),
            )
        else:
            d = np.linspace(-1 * self.infield_width, 0, 3)

            s, d = np.meshgrid(s, d)
            s, d = s.flatten(), d.flatten()
            tri = matplotlib.tri.Triangulation(s, d)

            x, y, z = self._transform_xyz(s, d)

            self._ax.plot_trisurf(
                x,
                y,
                z,
                triangles=tri.triangles,
                alpha=self.infield_alpha,
                color=self.infield_color,
                edgecolor=self.infield_color,
                zorder=-2,
            )

    def _draw_lanes(self, s_points=250) -> List:
        all_s = np.linspace(0, self.length, s_points)

        lane_positions = [0]
        for w in self.lane_widths:
            lane_positions.append(lane_positions[-1] + w)

        for d, color in zip(lane_positions, self.lane_colors):
            line = np.array([self._transform_xyz(s, d) for s in all_s])
            self._ax.plot(
                line[:, 0],
                line[:, 1],
                line[:, 2],
                color=color,
                alpha=self.lane_alpha,
                linestyle=self.lane_linestyle,
                lw=self.lane_linewidth,
                zorder=-1,
            )

    def _draw_lines(self) -> List:
        for s, color in zip(self.line_distances, self.line_colors):
            line = np.array([self._transform_xyz(s, d) for d in [0, self.width]])
            self._ax.plot(
                line[:, 0],
                line[:, 1],
                line[:, 2],
                color=color,
                alpha=self.line_alpha,
                zorder=-1,
            )

    def trajectory(
        self,
        s_: np.ndarray,
        d: np.ndarray,
        *args,
        **kwargs,
    ) -> matplotlib.lines.Line2D:
        """Plot a trajectory on the stadium.

        Parameters
        ----------
        s : np.ndarray
            The tangential position of the trajectory.
        d : np.ndarray
            The radial of the trajectory.
        args : list
            Additional positional arguments to pass to the plot
            function.
        kwargs : dict
            Additional keyword arguments to pass to the plot
            function.

        """
        points = np.array([self._transform_xyz(s_i, d_i) for s_i, d_i in zip(s_, d)])

        return self._ax.plot(points[:, 0], points[:, 1], points[:, 2], *args, **kwargs)

    def scatter(
        self,
        s_: np.ndarray,
        d_: np.ndarray,
        *args,
        **kwargs,
    ) -> matplotlib.collections.PathCollection:
        """Scatter points on the stadium.

        Parameters
        ----------
        s : np.ndarray
            The tangential position of the points.
        d : np.ndarray
            The radial of the points.
        args : list
            Additional positional arguments to pass to the scatter
            function.
        kwargs : dict
            Additional keyword arguments to pass to the scatter
            function.

        """
        points = np.array([self._transform_xyz(s_i, d_i) for s_i, d_i in zip(s_, d_)])

        return self._ax.scatter(points[:, 0], points[:, 1], points[:, 2], *args, **kwargs)
